from dataclasses import dataclass

from dagster_shared.record import record


def test_jit_check_collision():
    @record
    class Spec:
        name: list[str]

    @dataclass
    class LeakingClass:
        name: str

    checks = [LeakingClass("dont"), LeakingClass("leak"), LeakingClass("pls")]
    specs = []

    for check in checks:
        specs.append(
            Spec(
                name=[check.name],
            )
        )


def test_check_arg():
    @record
    class Test:
        check: list[str]

    arg = ["a", "b"]
    assert Test(check=arg).check == arg



import datetime
import os
from pathlib import Path

import dagster_shared.check as check
import pytest
import yaml
from dagster_shared.yaml_utils import (
    dump_run_config_yaml,
    load_run_config_yaml,
    load_yaml_from_glob_list,
    load_yaml_from_globs,
    load_yaml_from_path,
    merge_yaml_strings,
    merge_yamls,
)


def file_relative_path(file, path):
    return os.fspath((Path(file).parent / path).resolve())


def test_load_yaml():
    assert load_yaml_from_path(file_relative_path(__file__, "yamls/yaml_one.yaml")) == {
        "key_one": {"key_one_one": "value_one"}
    }


def test_from_glob_list():
    assert load_yaml_from_glob_list([file_relative_path(__file__, "yamls/yaml_one.yaml")]) == {
        "key_one": {"key_one_one": "value_one"}
    }

    assert load_yaml_from_glob_list(
        [
            file_relative_path(__file__, "yamls/yaml_one.yaml"),
            file_relative_path(__file__, "yamls/yaml_two.yaml"),
        ]
    ) == {"key_one": {"key_one_one": "value_one", "key_one_two": "value_two"}}

    assert load_yaml_from_glob_list([file_relative_path(__file__, "yamls/*.yaml")]) == {
        "key_one": {"key_one_one": "value_one", "key_one_two": "value_two"}
    }

    assert load_yaml_from_globs(
        file_relative_path(__file__, "yamls/yaml_one.yaml"),
        file_relative_path(__file__, "yamls/yaml_two.yaml"),
    ) == {"key_one": {"key_one_one": "value_one", "key_one_two": "value_two"}}

    assert load_yaml_from_glob_list(["flskhfhjsdf"]) == {}


def test_merge_yamls():
    assert merge_yamls(
        [
            file_relative_path(__file__, os.path.join("yamls", "yaml_one.yaml")),
            file_relative_path(__file__, os.path.join("yamls", "yaml_two.yaml")),
        ]
    ) == {"key_one": {"key_one_one": "value_one", "key_one_two": "value_two"}}

    with pytest.raises(
        check.CheckError,
        match=(
            r"Expected YAML from file .* to parse to dictionary, "
            'instead got: "this is a valid YAML string but not a dictionary"'
        ),
    ):
        merge_yamls(
            [
                file_relative_path(__file__, os.path.join("yamls", "yaml_one.yaml")),
                file_relative_path(__file__, os.path.join("yamls", "bad", "a_string.yaml")),
            ]
        )


def test_merge_yaml_strings():
    a = """
foo:
  bar: 1
baz: 3
"""
    b = """
foo:
  one: "one"
other: 4
"""
    c = """
final: "result"
"""

    assert merge_yaml_strings([a, b, c]) == {
        "final": "result",
        "foo": {"one": "one", "bar": 1},
        "other": 4,
        "baz": 3,
    }

    override = '''final: "some other result"'''
    assert merge_yaml_strings([a, b, c, override]) == {
        "foo": {"bar": 1, "one": "one"},
        "baz": 3,
        "other": 4,
        "final": "some other result",
    }

    string_yaml = "this is a valid YAML string but not a dictionary"
    expected = f'Expected YAML dictionary, instead got: "{string_yaml}"'

    with pytest.raises(check.CheckError, match=expected):
        merge_yaml_strings([a, string_yaml])

    with pytest.raises(
        yaml.YAMLError,
        match="while scanning for the next token\nfound character '`' that cannot start any token",
    ):
        bad_yaml = "--- `"
        merge_yaml_strings([a, bad_yaml])


def test_dump_octal_string():
    octal_str_list = {"keys": ["0001823", "0001234"]}

    # normal dump parses the first string as an int
    assert yaml.safe_dump(octal_str_list) == "keys:\n- 0001823\n- '0001234'\n"

    # our dump does not
    assert dump_run_config_yaml(octal_str_list) == "keys:\n- '0001823'\n- '0001234'\n"


def test_load_datetime_string():
    date_config_yaml = """ops:
  my_op:
    config:
      start: 2022-06-10T00:00:00.000000+00:00"""

    # normal dump parses as a datetime
    assert yaml.safe_load(date_config_yaml) == {
        "ops": {
            "my_op": {
                "config": {
                    "start": datetime.datetime(2022, 6, 10, 0, 0, tzinfo=datetime.timezone.utc)
                }
            }
        }
    }

    # ours does not
    assert load_run_config_yaml(date_config_yaml) == {
        "ops": {"my_op": {"config": {"start": "2022-06-10T00:00:00.000000+00:00"}}}
    }



from typing import TYPE_CHECKING, Annotated

import pytest
from dagster_shared.check import CheckError
from dagster_shared.check.builder import ImportFrom
from dagster_shared.check.decorator import checked

if TYPE_CHECKING:
    from dagster_shared.utils.test import TestType


def test_basic():
    @checked
    def foo(): ...

    foo()

    @checked
    def bar(i: int): ...

    bar(1)
    bar(i=1)
    with pytest.raises(CheckError):
        bar("1")  # type: ignore
    with pytest.raises(CheckError):
        bar(i="1")  # type: ignore


class Thing: ...


def test_many():
    @checked
    def big(
        name: str,
        nick_names: list[str],
        age: int,
        cool: bool,
        thing: Thing | None,
        other_thing: Thing,
        percent: float,
        o_s: str | None,
        o_n: int | None,
        o_f: float | None,
        o_b: bool | None,
        foos: list[Annotated["TestType", ImportFrom("dagster_shared.utils.test")]],
    ):
        return True

    assert big(
        name="dude",
        nick_names=[
            "foo",
            "bar",
            "biz",
        ],
        age=42,
        cool=False,
        thing=None,
        other_thing=Thing(),
        percent=0.5,
        o_s="x",
        o_n=3,
        o_f=None,
        o_b=None,
        foos=[],
    )

    with pytest.raises(CheckError):
        assert big(
            name="dude",
            nick_names=[
                "foo",
                "bar",
                "biz",
            ],
            age=42,
            cool=False,
            thing=None,
            other_thing=Thing(),
            percent=0.5,
            o_s="x",
            o_n=3,
            o_f="surprise_not_float",  # type: ignore
            o_b=None,
            foos=[],
        )


def test_no_op():
    def foo(): ...

    c_foo = checked(foo)
    assert c_foo is foo

    def bar() -> None: ...

    c_bar = checked(bar)
    assert c_bar is bar


def test_star():
    @checked
    def baz(*, i: int): ...

    baz(i=1)
    with pytest.raises(CheckError):
        baz(i="1")  # type: ignore


def test_partial():
    @checked
    def foo(a, b, c: int): ...

    foo(1, 2, 3)


def test_class():
    class Foo:
        @checked
        def me(self):
            return self

        @checked
        def yell(self, word: str):
            return word

        @staticmethod
        @checked
        def holler(word: str):
            return word

        @classmethod
        @checked
        def scream(cls, word: str):
            return word

    f = Foo()
    f.me()

    f.yell("HI")
    with pytest.raises(CheckError):
        f.yell(3)  # type: ignore

    Foo.holler("hi")
    with pytest.raises(CheckError):
        Foo.holler(3)  # type: ignore

    Foo.scream("hi")
    with pytest.raises(CheckError):
        Foo.scream(3)  # type: ignore


def test_defaults():
    @checked
    def foo(a: int, b: int = 1):
        return a + b

    assert foo(0) == 1

    @checked
    def bar(private_list: list = []):
        private_list.append(1)
        return private_list

    assert bar() == [1]
    assert bar() == [1]  # @checked makes it a not shared container instance

    class GlobalThing:
        def __init__(self):
            self._things = []

        def bump(self):
            self._things.append(1)

        def total(self):
            return sum(self._things)

    _global = GlobalThing()

    @checked
    def baz(shared_inst: GlobalThing = _global):
        _global.bump()
        return _global.total()

    assert baz() == 1
    assert baz() == 2



import datetime
import json
import logging
import os
import platform
import sys
import uuid
from collections.abc import Callable, Mapping
from logging.handlers import RotatingFileHandler
from typing import NamedTuple

import click

from dagster_shared.version import __version__

MAX_BYTES = 10485760  # 10 MB = 10 * 1024 * 1024 bytes

DAGSTER_HOME_FALLBACK = "~/.dagster"
TELEMETRY_STR = ".telemetry"
INSTANCE_ID_STR = "instance_id"
USER_ID_STR = "user_id"

KNOWN_CI_ENV_VAR_KEYS = {
    "GITLAB_CI",  # https://docs.gitlab.com/ee/ci/variables/predefined_variables.html
    "GITHUB_ACTION",  # https://docs.github.com/en/actions/learn-github-actions/variables#default-environment-variables
    "BITBUCKET_BUILD_NUMBER",  # https://support.atlassian.com/bitbucket-cloud/docs/variables-and-secrets/
    "JENKINS_URL",  # https://www.jenkins.io/doc/book/pipeline/jenkinsfile/#using-environment-variables
    "CODEBUILD_BUILD_ID",  # https://docs.aws.amazon.com/codebuild/latest/userguide/build-env-ref-env-vars.html
    "CIRCLECI",  # https://circleci.com/docs/variables/#built-in-environment-variables
    "TRAVIS",  # https://docs.travis-ci.com/user/environment-variables/#default-environment-variables
    "BUILDKITE",  # https://buildkite.com/docs/pipelines/environment-variables
}


def get_python_version() -> str:
    version = sys.version_info
    return f"{version.major}.{version.minor}.{version.micro}"


def get_is_known_ci_env() -> bool:
    # Many CI tools will use `CI` key which lets us know for sure it's a CI env
    if os.environ.get("CI"):
        return True

    # Otherwise looking for predefined env var keys of known CI tools
    for env_var_key in KNOWN_CI_ENV_VAR_KEYS:
        if env_var_key in os.environ:
            return True

    return False


def dagster_home_if_set() -> str | None:
    dagster_home_path = os.getenv("DAGSTER_HOME")

    if not dagster_home_path:
        return None

    return os.path.expanduser(dagster_home_path)


def get_or_create_dir_from_dagster_home(target_dir: str) -> str:
    """If $DAGSTER_HOME is set, return $DAGSTER_HOME/<target_dir>/
    Otherwise, return ~/.dagster/<target_dir>/.

    The 'logs' directory is used to cache logs before upload

    The '.logs_queue' directory is used to temporarily store logs during upload. This is to prevent
    dropping events or double-sending events that occur during the upload process.

    The '.telemetry' directory is used to store the instance id.
    """
    dagster_home_path = dagster_home_if_set()
    if dagster_home_path is None:
        dagster_home_path = os.path.expanduser(DAGSTER_HOME_FALLBACK)

    dagster_home_logs_path = os.path.join(dagster_home_path, target_dir)
    if not os.path.exists(dagster_home_logs_path):
        try:
            os.makedirs(dagster_home_logs_path)
        except FileExistsError:
            # let FileExistsError pass to avoid race condition when multiple places on the same
            # machine try to create this dir
            pass
    return dagster_home_logs_path


def get_or_create_user_telemetry_dir() -> str:
    """Always return ~/.dagster/.telemetry/, ignoring $DAGSTER_HOME.

    This is used for user-level telemetry data (like user_id) that should be
    consistent across all Dagster instances for a given user.
    """
    path = os.path.join(os.path.expanduser(DAGSTER_HOME_FALLBACK), TELEMETRY_STR)
    os.makedirs(path, exist_ok=True)
    return path


class TelemetrySettings(NamedTuple):
    dagster_telemetry_enabled: bool
    instance_id: str | None
    run_storage_id: str | None


def _get_rotating_file_handler(logger: logging.Logger) -> RotatingFileHandler | None:
    return next(
        iter(handler for handler in logger.handlers if isinstance(handler, RotatingFileHandler)),
        None,
    )


def get_telemetry_logger() -> logging.Logger:
    # If a concurrently running process deleted the logging directory since the
    # last action, we need to make sure to re-create the directory
    # (the logger does not do this itself.)
    dagster_home_path = get_or_create_dir_from_dagster_home("logs")

    logging_file_path = os.path.join(dagster_home_path, "event.log")
    logger = logging.getLogger("dagster_telemetry_logger")

    # Don't propagate to the root logger
    logger.propagate = False

    # If the file we were writing to has been overwritten, dump the existing handler and re-open the stream.
    if not os.path.exists(logging_file_path):
        handler = _get_rotating_file_handler(logger)
        if handler:
            handler.close()
            logger.removeHandler(handler)

    if not _get_rotating_file_handler(logger):
        handler = RotatingFileHandler(
            logging_file_path,
            maxBytes=MAX_BYTES,
            backupCount=10,
        )
        logger.setLevel(logging.INFO)
        logger.addHandler(handler)

    return logger


def write_telemetry_log_line(log_line: object) -> None:
    logger = get_telemetry_logger()
    logger.info(json.dumps(log_line))


# For use in test teardown
def cleanup_telemetry_logger() -> None:
    logger = logging.getLogger("dagster_telemetry_logger")

    handler = _get_rotating_file_handler(logger)
    if not handler:
        return

    handler.close()
    logger.removeHandler(handler)


class TelemetryEntry(
    NamedTuple(
        "_TelemetryEntry",
        [
            ("action", str),
            ("client_time", str),
            ("event_id", str),
            ("elapsed_time", str),
            ("instance_id", str),
            ("user_id", str),
            ("metadata", Mapping[str, str]),
            ("python_version", str),
            ("dagster_version", str),
            ("os_desc", str),
            ("os_platform", str),
            ("run_storage_id", str),
            ("is_known_ci_env", bool),
        ],
    )
):
    """Schema for telemetry logs.

    Currently, log entries are coerced to the same schema to enable storing all entries in one DB
    table with unified schema.

    action - Name of function called i.e. `execute_job_started` (see: fn telemetry_wrapper)
    client_time - Client time
    elapsed_time - Time elapsed between start of function and end of function call
    event_id - Unique id for the event
    instance_id - Unique id for dagster instance (varies by $DAGSTER_HOME)
    user_id - Unique id for the user (consistent across all instances, stored in ~/.dagster)
    python_version - Python version
    metadata - More information i.e. pipeline success (boolean)
    version - Schema version
    dagster_version - Version of the project being used.
    os_desc - String describing OS in use
    os_platform - Terse string describing OS platform - linux, windows, darwin, etc.
    run_storage_id - Unique identifier of run storage database

    If $DAGSTER_HOME is set, then use $DAGSTER_HOME/logs/
    Otherwise, use ~/.dagster/logs/
    """

    def __new__(
        cls,
        action: str,
        client_time: str,
        event_id: str,
        instance_id: str,
        user_id: str,
        metadata: Mapping[str, str] | None = None,
        elapsed_time: str | None = None,
        run_storage_id: str | None = None,
    ):
        OS_DESC = platform.platform()
        OS_PLATFORM = platform.system()

        return super().__new__(
            cls,
            action=action,
            client_time=client_time,
            elapsed_time=elapsed_time or "",
            event_id=event_id,
            instance_id=instance_id,
            user_id=user_id,
            python_version=get_python_version(),
            metadata=metadata or {},
            dagster_version=__version__ or "None",
            os_desc=OS_DESC,
            os_platform=OS_PLATFORM,
            run_storage_id=run_storage_id or "",
            is_known_ci_env=get_is_known_ci_env(),
        )


def log_telemetry_action(
    get_telemetry_settings: Callable[[], TelemetrySettings],
    action: str,
    client_time: datetime.datetime | None = None,
    elapsed_time: datetime.timedelta | None = None,
    metadata: Mapping[str, str] | None = None,
) -> None:
    if client_time is None:
        client_time = datetime.datetime.now()

    (dagster_telemetry_enabled, instance_id, run_storage_id) = get_telemetry_settings()

    if dagster_telemetry_enabled:
        if not instance_id:
            raise Exception("Instance ID must be set when telemetry is enabled")
        # Log general statistics
        write_telemetry_log_line(
            TelemetryEntry(
                action=action,
                client_time=str(client_time),
                elapsed_time=str(elapsed_time),
                event_id=str(uuid.uuid4()),
                instance_id=instance_id,
                user_id=get_or_set_user_id(),
                metadata=metadata,
                run_storage_id=run_storage_id,
            )._asdict()
        )


def get_telemetry_enabled_from_dagster_yaml() -> bool:
    """Lightweight check to see if telemetry is enabled by checking $DAGSTER_HOME/dagster.yaml,
    without needing to load the entire Dagster instance.
    """
    import yaml

    dagster_home_path = dagster_home_if_set()
    if dagster_home_path is None:
        return True

    dagster_yaml_path = os.path.join(dagster_home_path, "dagster.yaml")
    if not os.path.exists(dagster_yaml_path):
        return True

    with open(dagster_yaml_path, encoding="utf8") as dagster_yaml_file:
        dagster_yaml_data = yaml.safe_load(dagster_yaml_file)
        if (
            dagster_yaml_data
            and "telemetry" in dagster_yaml_data
            and "enabled" in dagster_yaml_data["telemetry"]
        ):
            return dagster_yaml_data["telemetry"]["enabled"]
    return True


def get_or_set_instance_id() -> str:
    instance_id = _get_telemetry_instance_id()
    if instance_id is None:
        instance_id = _set_telemetry_instance_id()
    return instance_id


def get_or_set_user_id() -> str:
    """Get or create a user-level telemetry ID.

    Unlike instance_id which varies by $DAGSTER_HOME, user_id is always stored
    in ~/.dagster/.telemetry/user_id.yaml to ensure consistency across all
    Dagster instances for a given user.
    """
    user_id = _get_telemetry_user_id()
    if user_id is None:
        user_id = _set_telemetry_user_id() or "<<unable_to_set_user_id>>"
    return user_id


def _get_telemetry_user_id() -> str | None:
    """Gets the user_id from ~/.dagster/.telemetry/user_id.yaml."""
    import yaml

    user_id_path = os.path.join(get_or_create_user_telemetry_dir(), "user_id.yaml")
    if not os.path.exists(user_id_path):
        return None

    try:
        with open(user_id_path, encoding="utf8") as user_id_file:
            user_id_yaml = yaml.safe_load(user_id_file)
            if (
                user_id_yaml
                and USER_ID_STR in user_id_yaml
                and isinstance(user_id_yaml[USER_ID_STR], str)
            ):
                return user_id_yaml[USER_ID_STR]
    except Exception:
        pass
    return None


def _set_telemetry_user_id() -> str | None:
    """Sets the user_id at ~/.dagster/.telemetry/user_id.yaml."""
    import yaml

    user_id_path = os.path.join(get_or_create_user_telemetry_dir(), "user_id.yaml")
    user_id = str(uuid.uuid4())

    try:
        with open(user_id_path, "w", encoding="utf8") as user_id_file:
            yaml.dump({USER_ID_STR: user_id}, user_id_file, default_flow_style=False)
        return user_id
    except Exception:
        return None


# Gets the instance_id at $DAGSTER_HOME/.telemetry/id.yaml
def _get_telemetry_instance_id() -> str | None:
    import yaml

    telemetry_id_path = os.path.join(get_or_create_dir_from_dagster_home(TELEMETRY_STR), "id.yaml")
    if not os.path.exists(telemetry_id_path):
        return

    with open(telemetry_id_path, encoding="utf8") as telemetry_id_file:
        telemetry_id_yaml = yaml.safe_load(telemetry_id_file)
        if (
            telemetry_id_yaml
            and INSTANCE_ID_STR in telemetry_id_yaml
            and isinstance(telemetry_id_yaml[INSTANCE_ID_STR], str)
        ):
            return telemetry_id_yaml[INSTANCE_ID_STR]
    return None


# Sets the instance_id at $DAGSTER_HOME/.telemetry/id.yaml
def _set_telemetry_instance_id() -> str:
    import yaml

    click.secho(TELEMETRY_TEXT, err=True)
    click.secho(SLACK_PROMPT, err=True)

    telemetry_id_path = os.path.join(get_or_create_dir_from_dagster_home(TELEMETRY_STR), "id.yaml")
    instance_id = str(uuid.uuid4())

    try:  # In case we encounter an error while writing to user's file system
        with open(telemetry_id_path, "w", encoding="utf8") as telemetry_id_file:
            yaml.dump({INSTANCE_ID_STR: instance_id}, telemetry_id_file, default_flow_style=False)
        return instance_id
    except Exception:
        return "<<unable_to_write_instance_id>>"


TELEMETRY_TEXT = """
  {telemetry}

  As an open-source project, we collect usage statistics to inform development priorities. For more
  information, read https://docs.dagster.io/about/telemetry.

  We will not see or store any data that is processed by your code.

  To opt-out, add the following to $DAGSTER_HOME/dagster.yaml, creating that file if necessary:

    telemetry:
      enabled: false
""".format(telemetry=click.style("Telemetry:", fg="blue", bold=True))

SLACK_PROMPT = """
  {welcome}

  If you have any questions or would like to engage with the Dagster team, please join us on Slack
  (https://bit.ly/39dvSsF).
""".format(welcome=click.style("Welcome to Dagster!", bold=True))



from dataclasses import dataclass

import pytest
from dagster_shared.seven import resolve_module_pattern


@pytest.mark.parametrize("matches_exist", [True, False])
def test_resolve_module_pattern_literal(monkeypatch, matches_exist: bool):
    def fake_find_spec(name):
        if name == "foo":
            return _FakeModuleSpec(["/fake/path/foo"])
        elif matches_exist and name == "foo.bar":
            return _FakeModuleSpec()
        return None

    def fake_iter_modules(path: list[str] | None = None):
        if path is None:
            return [
                (None, "foo", False),
            ]
        elif matches_exist and path == ["/fake/path/foo"]:
            return [(None, "bar", False)]
        return []

    monkeypatch.setattr("dagster_shared.seven.importlib.util.find_spec", fake_find_spec)
    monkeypatch.setattr("dagster_shared.seven.pkgutil.iter_modules", fake_iter_modules)

    result = resolve_module_pattern("foo.bar")
    if matches_exist:
        assert result == ["foo.bar"]
    else:
        assert result == []


def test_resolve_module_pattern_top_level_standalone_wildcard(monkeypatch):
    def fake_find_spec(name: str):
        return _FakeModuleSpec()

    def fake_iter_modules(path: list[str] | None = None):
        return [
            (None, "foo", False),
            (None, "bar", False),
            (None, "baz", False),
        ]

    monkeypatch.setattr("dagster_shared.seven.importlib.util.find_spec", fake_find_spec)
    monkeypatch.setattr("dagster_shared.seven.pkgutil.iter_modules", fake_iter_modules)

    monkeypatch.setattr(
        "dagster_shared.seven.importlib.util.find_spec",
        fake_find_spec,
    )

    result = resolve_module_pattern("*")

    assert set(result) == {"foo", "bar", "baz"}


@pytest.mark.parametrize("matches_exist", [True, False])
def test_resolve_module_pattern_embedded_wildcard_start(monkeypatch, matches_exist: bool):
    def fake_find_spec(name):
        if name in ["bar", "baz"]:
            return _FakeModuleSpec([f"/fake/path/{name}"])
        if matches_exist and name in {"bar.foo", "baz.foo"}:
            return _FakeModuleSpec()
        return None

    def fake_iter_modules(path: list[str] | None = None):
        if matches_exist and path == ["/fake/path/bar"]:
            return [(None, "foo", False)]
        elif matches_exist and path == ["/fake/path/bar"]:
            return [(None, "foo", False)]
        return [
            (None, "bar", False),
            (None, "baz", False),
        ]

    monkeypatch.setattr("dagster_shared.seven.importlib.util.find_spec", fake_find_spec)
    monkeypatch.setattr("dagster_shared.seven.pkgutil.iter_modules", fake_iter_modules)

    result = resolve_module_pattern("*.foo")
    if matches_exist:
        assert result == ["bar.foo", "baz.foo"]
    else:
        assert result == []


@pytest.mark.parametrize("matches_exist", [True, False])
def test_resolve_wildcard_modules_embedded_wildcard_middle(monkeypatch, matches_exist: bool):
    def fake_find_spec(name: str) -> _FakeModuleSpec | None:
        if name == "foo":
            return _FakeModuleSpec(["/fake/path/foo"])
        if matches_exist and name in ["foo.bar.models", "foo.baz.models"]:
            return _FakeModuleSpec()
        return None

    def fake_iter_modules(path: list[str] | None = None):
        if path == ["/fake/path/foo"]:
            return [
                (None, "bar", False),
                (None, "baz", False),
            ]
        return []

    monkeypatch.setattr("dagster_shared.seven.importlib.util.find_spec", fake_find_spec)
    monkeypatch.setattr("dagster_shared.seven.pkgutil.iter_modules", fake_iter_modules)

    result = resolve_module_pattern("foo.*.models")
    if matches_exist:
        assert result == ["foo.bar.models", "foo.baz.models"]
    else:
        assert result == []


@pytest.mark.parametrize("matches_exist", [True, False])
def test_resolve_module_pattern_embedded_wildcard_end(monkeypatch, matches_exist: bool):
    def fake_find_spec(name):
        if name == "foo":
            return _FakeModuleSpec(["/fake/path/foo"])
        if matches_exist and name in {"foo.bar", "foo.baz"}:
            return _FakeModuleSpec()
        return None

    def fake_iter_modules(path: list[str] | None = None):
        if path == ["/fake/path/foo"]:
            return [
                (None, "bar", False),
                (None, "baz", False),
            ]
        return []

    monkeypatch.setattr("dagster_shared.seven.importlib.util.find_spec", fake_find_spec)
    monkeypatch.setattr("dagster_shared.seven.pkgutil.iter_modules", fake_iter_modules)

    result = resolve_module_pattern("foo.*")
    if matches_exist:
        assert result == ["foo.bar", "foo.baz"]
    else:
        assert result == []


@dataclass
class _FakeModuleSpec:
    submodule_search_locations: list[str] | None = None



from dagster_shared.serdes.objects.package_entry import (
    ComponentFeatureData as ComponentFeatureData,
    EnvRegistryKey as EnvRegistryKey,
    EnvRegistryObjectSnap as EnvRegistryObjectSnap,
    ScaffoldTargetTypeData as ScaffoldTargetTypeData,
)



from collections.abc import Sequence
from typing import Any, Literal, NamedTuple, Union

import pytest
from dagster_shared.match import match_type


# Helper to simulate the is_named_tuple_instance logic used in match_type
class MyTuple(NamedTuple):
    x: int
    y: str


@pytest.mark.parametrize(
    "obj,type_,expected",
    [
        # Concrete types
        (5, int, True),
        ("hello", str, True),
        (None, type(None), True),
        (5.5, int, False),
        (5.5, Any, True),
        ("abc", Any, True),
        ([1], Any, True),
        # Union
        (5, Union[int, str], True),  # noqa: UP007
        ("hi", Union[int, str], True),  # noqa: UP007
        (5.0, Union[int, str], False),  # noqa: UP007
        # Union via pipe syntax (types.UnionType)
        (5, int | str, True),
        ("hi", int | str, True),
        (5.0, int | str, False),
        # Literal
        ("hello", Literal["hello", "world"], True),
        ("nope", Literal["hello", "world"], False),
        (3, Literal[3, 5], True),
        (4, Literal[3, 5], False),
        # list[]
        ([1, 2, 3], list[int], True),
        ([1, "x"], list[int], False),
        ([], list[int], True),
        # Sequence[] support
        ([1, 2], Sequence[int], True),
        ((1, 2), Sequence[int], True),
        ({1, 2}, Sequence[int], False),
        # tuple[...] fixed-length
        ((1, "x"), tuple[int, str], True),
        ((1, 2), tuple[int, str], False),
        ((1,), tuple[int, str], False),
        # tuple[...] variable-length
        ((1, 2, 3), tuple[int, ...], True),
        ((1, "x"), tuple[int, ...], False),
        # dict[]
        ({"a": 1, "b": 2}, dict[str, int], True),
        ({"a": 1, 2: "b"}, dict[str, int], False),
        ({}, dict[str, int], True),
        # NamedTuple
        (MyTuple(1, "a"), NamedTuple, True),
        ((1, "a"), NamedTuple, False),
        # tuple of types
        ("abc", (int, str), True),
        ("abc", (int, Literal["abc", "def"]), True),
        ("ghi", (int, Literal["abc", "def"]), False),
    ],
)
def test_match_type(obj, type_, expected):
    assert match_type(obj, type_) == expected



from collections.abc import Mapping, Sequence
from typing import Any

from dagster_shared.record import as_dict, record


@record
class DgAssetMetadata:
    asset_key: str
    dependency_keys: list[str]
    kinds: list[str]
    group_name: str | None
    description: str | None
    automation_condition: Mapping[str, Any] | None
    tags: Sequence[Mapping[str, str]]
    is_executable: bool
    source: str | None
    owners: Sequence[Mapping[str, str]] | None


@record
class DgSensorMetadata:
    name: str
    source: str | None


@record
class DgScheduleMetadata:
    name: str
    cron_schedule: str
    source: str | None


@record
class DgJobMetadata:
    name: str
    description: str | None
    source: str | None


@record
class DgResourceMetadata:
    name: str
    type: str


@record
class DgAssetCheckMetadata:
    key: str
    asset_key: str
    name: str
    additional_deps: list[str]
    description: str | None
    source: str | None


@record
class DgDefinitionMetadata:
    assets: list[DgAssetMetadata]
    asset_checks: list[DgAssetCheckMetadata]
    jobs: list[DgJobMetadata]
    resources: list[DgResourceMetadata]
    schedules: list[DgScheduleMetadata]
    sensors: list[DgSensorMetadata]

    def to_dict(self) -> Mapping[str, Sequence[Mapping[str, Any]]]:
        return {
            "assets": [as_dict(asset) for asset in self.assets],
            "asset_checks": [as_dict(asset_check) for asset_check in self.asset_checks],
            "jobs": [as_dict(job) for job in self.jobs],
            "resources": [as_dict(resource) for resource in self.resources],
            "schedules": [as_dict(schedule) for schedule in self.schedules],
            "sensors": [as_dict(sensor) for sensor in self.sensors],
        }

    @property
    def is_empty(self) -> bool:
        return (
            not self.assets
            and not self.asset_checks
            and not self.jobs
            and not self.resources
            and not self.schedules
            and not self.sensors
        )



import functools
import os

import click

from dagster_shared.plus.config import DagsterPlusCliConfig

# Constants for CLI arguments and environment variables
DEPLOYMENT_CLI_ARGUMENT = "deployment"
DEPLOYMENT_ENV_VAR_NAME = "DAGSTER_CLOUD_DEPLOYMENT"

ORGANIZATION_CLI_ARGUMENT = "organization"
ORGANIZATION_ENV_VAR_NAME = "DAGSTER_CLOUD_ORGANIZATION"

TOKEN_CLI_ARGUMENT = "api_token"
TOKEN_ENV_VAR_NAME = "DAGSTER_CLOUD_API_TOKEN"


def get_deployment(ctx: click.Context | None = None) -> str | None:
    """Gets the configured deployment to target.
    Highest precedence is a deployment argument, then `DAGSTER_CLOUD_DEPLOYMENT` env var, then config file default.
    """
    if ctx and ctx.params.get(DEPLOYMENT_CLI_ARGUMENT):
        return ctx.params[DEPLOYMENT_CLI_ARGUMENT]

    env_value = os.getenv(DEPLOYMENT_ENV_VAR_NAME)
    if env_value:
        return env_value

    # Fall back to config file
    if DagsterPlusCliConfig.exists():
        try:
            config = DagsterPlusCliConfig.get()
            return config.default_deployment
        except Exception:
            pass

    return None


def get_organization(ctx: click.Context | None = None) -> str | None:
    """Gets the configured organization to target.
    Highest precedence is an organization argument, then `DAGSTER_CLOUD_ORGANIZATION` env var, then config file.
    """
    if ctx and ctx.params.get(ORGANIZATION_CLI_ARGUMENT):
        return ctx.params[ORGANIZATION_CLI_ARGUMENT]

    env_value = os.getenv(ORGANIZATION_ENV_VAR_NAME)
    if env_value:
        return env_value

    # Fall back to config file
    if DagsterPlusCliConfig.exists():
        try:
            config = DagsterPlusCliConfig.get()
            return config.organization
        except Exception:
            pass

    return None


def get_user_token(ctx: click.Context | None = None) -> str | None:
    """Gets the configured user token to use.
    Highest precedence is an api-token argument, then `DAGSTER_CLOUD_API_TOKEN` env var, then config file.
    """
    if ctx and ctx.params.get(TOKEN_CLI_ARGUMENT):
        return ctx.params[TOKEN_CLI_ARGUMENT]

    env_value = os.getenv(TOKEN_ENV_VAR_NAME)
    if env_value:
        return env_value

    # Fall back to config file
    if DagsterPlusCliConfig.exists():
        try:
            config = DagsterPlusCliConfig.get()
            return config.user_token
        except Exception:
            pass

    return None


# Click option definitions for reuse across commands
DEPLOYMENT_OPTION = click.option(
    "--deployment",
    "-d",
    help="Deployment to target.",
    envvar=DEPLOYMENT_ENV_VAR_NAME,
)

ORGANIZATION_OPTION = click.option(
    "--organization",
    "-o",
    help="Organization to target.",
    envvar=ORGANIZATION_ENV_VAR_NAME,
)

TOKEN_OPTION = click.option(
    "--api-token",
    help="Dagster Cloud API token.",
    envvar=TOKEN_ENV_VAR_NAME,
)

VIEW_GRAPHQL_OPTION = click.option(
    "--view-graphql",
    is_flag=True,
    help="Print GraphQL queries and responses to stderr for debugging.",
)


def dg_api_options(
    deployment_scoped: bool = False,
    organization_scoped: bool = False,
):
    """Apply this decorator to Click commands to add organization, deployment, and token options.

    Args:
        deployment_scoped: If True, requires deployment to be specified and validates it
        organization_scoped: If True, only requires organization (deployment is optional)
    """
    if deployment_scoped and organization_scoped:
        raise ValueError("Cannot specify both deployment_scoped and organization_scoped")

    if not deployment_scoped and not organization_scoped:
        raise ValueError("Must specify either deployment_scoped or organization_scoped")

    def decorator(func):
        # Add options in reverse order (Click applies them in reverse)
        func = VIEW_GRAPHQL_OPTION(func)
        func = TOKEN_OPTION(func)
        func = ORGANIZATION_OPTION(func)

        if deployment_scoped:
            func = DEPLOYMENT_OPTION(func)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            ctx = click.get_current_context()

            # Check if we're in test mode with DgApiTestContext
            # Import here to avoid circular imports
            from dagster_dg_cli.cli.api.client import DgApiTestContext

            if ctx.obj and isinstance(ctx.obj, DgApiTestContext):
                organization = ctx.obj.organization
                deployment = ctx.obj.deployment if deployment_scoped else None
                api_token = None
                view_graphql = False  # Default to False in test mode
            else:
                # Get resolved values using precedence chain
                organization = get_organization(ctx)
                deployment = get_deployment(ctx) if deployment_scoped else None
                api_token = get_user_token(ctx)
                view_graphql = kwargs.get("view_graphql", False)

            if not organization:
                raise click.UsageError(
                    "A Dagster Cloud organization must be specified.\n\n"
                    "You may specify an organization by:\n"
                    f"- Providing the --organization parameter\n"
                    f"- Setting the {ORGANIZATION_ENV_VAR_NAME} environment variable"
                )

            if deployment_scoped and not deployment:
                raise click.UsageError(
                    "A Dagster Cloud deployment must be specified for this command.\n\n"
                    "You may specify a deployment by:\n"
                    f"- Providing the --deployment parameter\n"
                    f"- Setting the {DEPLOYMENT_ENV_VAR_NAME} environment variable"
                )

            # Update kwargs with resolved values
            kwargs["organization"] = organization
            kwargs["api_token"] = api_token
            kwargs["view_graphql"] = view_graphql
            if deployment_scoped:
                kwargs["deployment"] = deployment

            return func(*args, **kwargs)

        return wrapper

    return decorator



import copy
from collections.abc import Mapping
from typing import TypeVar, cast

import dagster_shared.check as check

K = TypeVar("K")
V = TypeVar("V")
K2 = TypeVar("K2")
V2 = TypeVar("V2")


def _deep_merge_dicts(
    onto_dict: dict[K, V], from_dict: Mapping[K2, V2]
) -> dict[K | K2, V | V2 | dict[object, object]]:
    check.mapping_param(from_dict, "from_dict")
    check.dict_param(onto_dict, "onto_dict")

    _onto_dict = cast("dict[K | K2, V | V2 | dict[object, object]]", onto_dict)
    for from_key, from_value in from_dict.items():
        if from_key not in onto_dict:
            _onto_dict[from_key] = from_value
        else:
            onto_value = _onto_dict[from_key]

            if isinstance(from_value, dict) and isinstance(onto_value, dict):
                _onto_dict[from_key] = _deep_merge_dicts(onto_value, from_value)
            else:
                _onto_dict[from_key] = from_value  # smash

    return _onto_dict


def deep_merge_dicts(
    onto_dict: Mapping[K, V], from_dict: Mapping[K2, V2]
) -> dict[K | K2, V | V2 | dict[object, object]]:
    """Returns a recursive union of two input dictionaries:
    * The returned dictionary has an entry for any key that's in either of the inputs.
    * For any key whose value is a dictionary in both of the inputs, the returned value will
      be the result of deep-merging the two input sub-dictionaries.

    If from_dict and onto_dict have different values for the same key, and the values are not both
    dictionaries, the returned dictionary contains the value from from_dict.
    """
    onto_dict = copy.deepcopy(onto_dict if isinstance(onto_dict, dict) else dict(onto_dict))
    return _deep_merge_dicts(onto_dict, from_dict)



from collections.abc import Callable, Sequence
from typing import Any, TypeAlias, TypeVar

import click
from typing_extensions import Self

from dagster_shared.record import as_dict, record

T_Callable = TypeVar("T_Callable", bound=Callable[..., Any])

ClickOption: TypeAlias = Callable[[T_Callable], T_Callable]


def apply_click_params(command: T_Callable, *click_params: ClickOption) -> T_Callable:
    for click_param in click_params:
        command = click_param(command)
    return command


def python_pointer_options(f: T_Callable) -> T_Callable:
    return apply_click_params(
        f,
        *generate_python_pointer_options(
            allow_multiple=False,
            hide_legacy_options=False,
        ),
    )


def generate_python_pointer_options(
    *,
    allow_multiple: bool,
    hide_legacy_options: bool,
) -> Sequence[ClickOption]:
    return [
        click.option(
            "--working-directory",
            "-d",
            help="Specify working directory to use when loading the repository or job",
            envvar="DAGSTER_WORKING_DIRECTORY",
        ),
        click.option(
            "--python-file",
            "-f",
            # Checks that the path actually exists lower in the stack, where we
            # are better equipped to surface errors
            type=click.Path(exists=False),
            multiple=allow_multiple,
            help=(
                "Specify python file "
                + ("or files (flag can be used multiple times) " if allow_multiple else "")
                + "where dagster definitions reside as top-level symbols/variables and load "
                + ("each" if allow_multiple else "the")
                + " file as a code location in the current python environment."
            ),
            envvar="DAGSTER_PYTHON_FILE",
        ),
        click.option(
            "--module-name",
            "-m",
            multiple=allow_multiple,
            help=(
                "Specify module "
                + ("or modules (flag can be used multiple times) " if allow_multiple else "")
                + "where dagster definitions reside as top-level symbols/variables and load "
                + ("each" if allow_multiple else "the")
                + " module as a code location in the current python environment."
            ),
            envvar="DAGSTER_MODULE_NAME",
        ),
        click.option(
            "--autoload-defs-module-name",
            help=("A module to import and recursively search through for definitions."),
            envvar="DAGSTER_autoload_defs_module_name",
        ),
        click.option(
            "--package-name",
            multiple=allow_multiple,
            help="Specify Python package where repository or job function lives",
            envvar="DAGSTER_PACKAGE_NAME",
            hidden=hide_legacy_options,
        ),
        click.option(
            "--attribute",
            "-a",
            help=(
                "Attribute that is either a 1) repository or job or "
                "2) a function that returns a repository or job"
            ),
            envvar="DAGSTER_ATTRIBUTE",
            hidden=hide_legacy_options,
        ),
    ]


def generate_grpc_server_options() -> Sequence[ClickOption]:
    return []


def generate_workspace_options(*, hide_uncommon_options: bool) -> Sequence[ClickOption]:
    return [
        click.option("--empty-workspace", is_flag=True, help="Allow an empty workspace"),
        click.option(
            "--workspace",
            "-w",
            multiple=True,
            type=click.Path(exists=True),
            help="Path to workspace file. Argument can be provided multiple times.",
        ),
        *generate_python_pointer_options(
            allow_multiple=True,
            hide_legacy_options=hide_uncommon_options,
        ),
        click.option(
            "--grpc-port",
            type=click.INT,
            required=False,
            help="Port to use to connect to gRPC server",
            hidden=hide_uncommon_options,
        ),
        click.option(
            "--grpc-socket",
            type=click.Path(),
            required=False,
            help="Named socket to use to connect to gRPC server",
            hidden=hide_uncommon_options,
        ),
        click.option(
            "--grpc-host",
            type=click.STRING,
            required=False,
            help="Host to use to connect to gRPC server, defaults to localhost",
            hidden=hide_uncommon_options,
        ),
        click.option(
            "--use-ssl",
            is_flag=True,
            default=False,
            help="Use a secure channel when connecting to the gRPC server",
            hidden=hide_uncommon_options,
        ),
    ]


def dg_workspace_options(f: T_Callable) -> T_Callable:
    return apply_click_params(
        f,
        *generate_workspace_options(
            hide_uncommon_options=True,  # dg as a newer cli hides deprecated and rarely used options
        ),
    )


def workspace_options(f: T_Callable) -> T_Callable:
    return apply_click_params(
        f,
        *generate_workspace_options(
            hide_uncommon_options=False,
        ),
    )


@record
class WorkspaceOpts:
    empty_workspace: bool = False
    workspace: Sequence[str] | None = None

    # Like PythonPointerParams but multiple files/modules/packages are allowed
    python_file: Sequence[str] | None = None
    module_name: Sequence[str] | None = None
    package_name: Sequence[str] | None = None
    working_directory: str | None = None
    attribute: str | None = None

    autoload_defs_module_name: str | None = None

    # For gRPC server
    grpc_port: int | None = None
    grpc_socket: str | None = None
    grpc_host: str | None = None
    use_ssl: bool = False

    @classmethod
    def extract_from_cli_options(cls, cli_options: dict[str, Any]) -> Self:
        # This is expected to always be called from a click entry point, so all options should be
        # present in the dictionary. We rely on `@record` for type-checking.
        return cls(
            empty_workspace=cli_options.pop("empty_workspace", False),
            workspace=cli_options.pop("workspace", None),
            python_file=cli_options.pop("python_file", None),
            module_name=cli_options.pop("module_name", None),
            package_name=cli_options.pop("package_name", None),
            working_directory=cli_options.pop("working_directory", None),
            attribute=cli_options.pop("attribute", None),
            autoload_defs_module_name=cli_options.pop("autoload_defs_module_name", None),
            grpc_port=cli_options.pop("grpc_port", None),
            grpc_socket=cli_options.pop("grpc_socket", None),
            grpc_host=cli_options.pop("grpc_host", None),
            use_ssl=cli_options.pop("use_ssl", False),
        )

    def specifies_target(self) -> bool:
        set_args = [
            k for k, v in as_dict(self).items() if v and (k not in {"empty_workspace", "use_ssl"})
        ]
        return bool(set_args)


@record
class PythonPointerOpts:
    python_file: str | None = None
    module_name: str | None = None
    package_name: str | None = None
    working_directory: str | None = None
    attribute: str | None = None
    autoload_defs_module_name: str | None = None

    @classmethod
    def extract_from_cli_options(cls, cli_options: dict[str, Any]) -> Self:
        # This is expected to always be called from a click entry point, so all options should be
        # present in the dictionary. We rely on `@record` for type-checking.
        return cls(
            python_file=cli_options.pop("python_file", None),
            module_name=cli_options.pop("module_name", None),
            package_name=cli_options.pop("package_name", None),
            working_directory=cli_options.pop("working_directory", None),
            attribute=cli_options.pop("attribute", None),
            autoload_defs_module_name=cli_options.pop("autoload_defs_module_name", None),
        )

    def to_workspace_opts(self) -> "WorkspaceOpts":
        return WorkspaceOpts(
            python_file=(self.python_file,) if self.python_file else None,
            module_name=(self.module_name,) if self.module_name else None,
            package_name=(self.package_name,) if self.package_name else None,
            autoload_defs_module_name=self.autoload_defs_module_name
            if self.autoload_defs_module_name
            else None,
            working_directory=self.working_directory,
            attribute=self.attribute,
        )

    def specifies_target(self) -> bool:
        return bool(self.python_file or self.module_name or self.package_name)



import json
import random
import string
from http.server import BaseHTTPRequestHandler, HTTPServer
from urllib import parse


def create_token_callback_handler(nonce: str) -> type[BaseHTTPRequestHandler]:
    class TokenCallbackHandler(BaseHTTPRequestHandler):
        def _send_shared_headers(self):
            self.send_header("Access-Control-Allow-Origin", "*")
            self.send_header("Access-Control-Allow-Methods", "POST,OPTIONS")
            self.send_header("Access-Control-Allow-Headers", "Content-Type")

        def _error(self, code, message=None):
            self.send_error(code, message)
            self._send_shared_headers()
            self.end_headers()

        def do_OPTIONS(self):
            self.send_response(200)
            self._send_shared_headers()
            self.end_headers()

        def do_POST(self):
            url = parse.urlparse(self.path)

            if url.path != "/callback":
                self._error(404, "Page not found")
                return

            content = self.rfile.read(int(self.headers["Content-Length"]))
            body = json.loads(content)

            if not body.get("nonce") or body.get("nonce") != nonce:
                self._error(400, "Invalid nonce")
                return
            if not body.get("token"):
                self._error(400, "No user token provided")
                return
            if not body.get("organization"):
                self._error(400, "No organization provided")
                return

            organization = body.get("organization")
            token = body.get("token")

            self.send_response(200)
            self._send_shared_headers()
            response = b'{ "ok": true }'
            self.send_header("Content-length", str(len(response)))
            self.end_headers()

            self.wfile.write(response)

            # Pass back result values
            assert isinstance(self.server, TokenServer)
            self.server.set_result(organization=organization, token=token)
            self.server.shutdown()

        # Overridden so that the webserver doesn't log requests to the console

        def log_request(self, code="-", size="-"):
            return

    return TokenCallbackHandler


class TokenServer(HTTPServer):
    organization: str | None = None
    token: str | None = None

    def __init__(self, host: tuple[str, int], nonce: str):
        super().__init__(host, create_token_callback_handler(nonce))

    def shutdown(self):
        # Stop serving the token server
        # https://stackoverflow.com/a/36017741
        setattr(self, "_BaseServer__shutdown_request", True)

    def set_result(self, organization: str, token: str):
        self.organization = organization
        self.token = token

    def get_organization(self) -> str | None:
        return self.organization

    def get_token(self) -> str | None:
        return self.token


def _generate_nonce():
    return "".join(
        random.SystemRandom().choice(string.ascii_uppercase + string.digits) for _ in range(8)
    )


def start_login_server(base_url: str | None = None) -> tuple[TokenServer, str]:
    """Starts a login server on a free port and returns
    the server and the URL to open in the browser.
    """
    from dagster_shared.utils import find_free_port

    port = find_free_port()
    nonce = _generate_nonce()
    escaped = parse.quote(f"/cli-auth/{nonce}?port={port}")
    auth_url = f"{base_url or 'https://dagster.cloud'}?next={escaped}"

    server = TokenServer(("localhost", port), nonce)
    return server, auth_url



import os
import time
from contextlib import contextmanager


def format_duration(milliseconds: float) -> str:
    """Given milliseconds, return human readable duration string such as:
    533ms, 2.1s, 4m52s, 34m12s, 1h4m.
    """
    # under 1 ms
    # ex: 0.83ms
    # ex: 8.3ms
    if milliseconds < 10:
        return f"{round(milliseconds, 2)}ms"

    # between 10 ms and 1000 ms
    # ex: 533ms
    if milliseconds < 1000:
        return f"{int(milliseconds)}ms"

    # between one second and one minute
    # ex: 5.6s
    if milliseconds < 1000 * 60:
        seconds = milliseconds / 1000
        return f"{round(seconds, 2)}s"

    # between one minute and 60 minutes
    # 5m42s
    if milliseconds < 1000 * 60 * 60:
        minutes = int(milliseconds // (1000 * 60))
        seconds = int(milliseconds % (1000 * 60) // (1000))
        return f"{minutes}m{seconds}s"

    # Above one hour
    else:
        hours = int(milliseconds // (1000 * 60 * 60))
        minutes = int(milliseconds % (1000 * 60 * 60) // (1000 * 60))
        return f"{hours}h{minutes}m"


@contextmanager
def fixed_timezone(tzname: str = "UTC"):
    """Temporarily set the process-wide timezone to `tzname`.
    Temporarily set the process-wide timezone to `tzname`.

    Example:
        with fixed_timezone("America/New_York"):
            dt = datetime.fromtimestamp(0)
            print(dt.strftime("%Y-%m-%d %H:%M %Z"))
            # 1969-12-31 19:00 EST
    """
    prev = os.environ.get("TZ")
    os.environ["TZ"] = tzname
    if hasattr(time, "tzset"):
        time.tzset()
    try:
        yield
    finally:
        if prev is None:
            os.environ.pop("TZ", None)
        else:
            os.environ["TZ"] = prev
        if hasattr(time, "tzset"):
            time.tzset()



# record depends on check, but check needs to discern records so these pieces are defined here

RECORD_MARKER_VALUE = object()
# "I do want to release this as checkrepublic one day" - schrockn
RECORD_MARKER_FIELD = "__checkrepublic__"


def is_record(obj) -> bool:
    """Whether or not this object was produced by a record decorator."""
    return getattr(obj, RECORD_MARKER_FIELD, None) == RECORD_MARKER_VALUE



import os
import pickle
from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from functools import cached_property
from typing import TYPE_CHECKING, Annotated, Any, Generic, NamedTuple, Optional, TypeVar

import pytest
from dagster_shared.check.builder import INJECTED_DEFAULT_VALS_LOCAL_VAR
from dagster_shared.check.functions import CheckError
from dagster_shared.record import (
    IHaveNew,
    ImportFrom,
    LegacyNamedTupleMixin,
    build_args_and_assignment_strs,
    check,
    copy,
    record,
    record_custom,
    replace,
)
from dagster_shared.serdes.serdes import deserialize_value, serialize_value, whitelist_for_serdes
from dagster_shared.utils.cached_method import cached_method
from dagster_shared.utils.hash import hash_collection
from pydantic import BaseModel, Field, TypeAdapter

if TYPE_CHECKING:
    from dagster_shared.utils.test import TestType


def test_kwargs_only() -> None:
    @record
    class MyClass:
        foo: str

    with pytest.raises(TypeError, match="takes 1 positional argument but 2 were given"):
        MyClass("fdslk")  # type: ignore # good job type checker


def test_runtime_typecheck() -> None:
    @record
    class MyClass:
        foo: str
        bar: int

    with pytest.raises(check.CheckError):
        MyClass(foo="fdsjk", bar="fdslk")  # type: ignore # good job type checker


def test_override_constructor_in_subclass() -> None:
    @record_custom
    class MyClass(IHaveNew):
        foo: str
        bar: int

        def __new__(cls, foo: str, bar: int):
            return super().__new__(
                cls,
                foo=foo,
                bar=bar,
            )

    assert MyClass(foo="fdsjk", bar=4)


def test_override_constructor_in_subclass_different_arg_names() -> None:
    @record_custom
    class MyClass(IHaveNew):
        foo: str
        bar: int

        def __new__(cls, fooarg: str, bararg: int):
            return super().__new__(
                cls,
                foo=fooarg,
                bar=bararg,
            )

    assert MyClass(fooarg="fdsjk", bararg=4)


def test_override_constructor_in_subclass_wrong_type() -> None:
    @record_custom
    class MyClass(IHaveNew):
        foo: str
        bar: int

        def __new__(cls, foo: str, bar: str):
            return super().__new__(
                cls,
                foo=foo,
                bar=bar,
            )

    with pytest.raises(check.CheckError):
        MyClass(foo="fdsjk", bar="fdslk")


def test_model_copy() -> None:
    @record
    class MyClass2:
        foo: str
        bar: int

    obj = MyClass2(foo="abc", bar=5)

    assert copy(obj, foo="xyz") == MyClass2(foo="xyz", bar=5)
    assert copy(obj, bar=6) == MyClass2(foo="abc", bar=6)
    assert copy(obj, foo="xyz", bar=6) == MyClass2(foo="xyz", bar=6)


def test_non_record_param():
    class SomeClass: ...

    class OtherClass: ...

    @record
    class MyModel2:
        some_class: SomeClass

    assert MyModel2(some_class=SomeClass())

    with pytest.raises(check.CheckError):
        MyModel2(some_class=OtherClass())  # wrong class  # pyright: ignore[reportArgumentType]

    with pytest.raises(check.CheckError):
        MyModel2(some_class=SomeClass)  # forgot ()  # pyright: ignore[reportArgumentType]


def test_cached_method() -> None:
    @record()
    class CoolModel2:
        name: str

        @cached_method
        def calculate(self, n: int):
            return {self.name: n}

        @cached_method
        def reticulate(self, n: int):
            return {self.name: n}

        @property
        @cached_method
        def prop(self):
            return {"four": 4}

    m = CoolModel2(name="bob")
    assert m.calculate(4) is m.calculate(4)
    assert m.calculate(4) is not m.reticulate(4)
    assert m.prop is m.prop

    clean_m = CoolModel2(name="bob")
    # cache doesn't effect equality
    assert clean_m == m
    assert m == clean_m

    # cache doesn't effect hash
    s = {m, clean_m}
    assert len(s) == 1

    # cache erased on copy
    m_copy = copy(m)
    assert m_copy.calculate(4) is not m.calculate(4)


def test_forward_ref() -> None:
    @record
    class Parent:
        partner: Optional["Parent"]
        child: Optional["Child"]

    class Child: ...

    assert Parent(partner=None, child=None)


def test_forward_ref_with_new() -> None:
    @record_custom
    class Parent:
        partner: Optional["Parent"]
        child: Optional["Child"]

        def __new__(cls, partner=None, child=None):
            return super().__new__(
                cls,
                partner=partner,  # pyright: ignore[reportCallIssue]
                child=child,  # pyright: ignore[reportCallIssue]
            )

    class Child: ...

    assert Parent()


def _empty_callsite_scope(cls, arg):
    cls(local=arg)


def test_frame_capture() -> None:
    class LocalAtDefineTime: ...

    val = LocalAtDefineTime()

    @record
    class Direct:
        local: Optional["LocalAtDefineTime"]

    _empty_callsite_scope(Direct, val)

    @record()  # invoking decorator has different frame depth from direct
    class InDirect:
        local: Optional["LocalAtDefineTime"]

    _empty_callsite_scope(InDirect, val)

    @record_custom
    class DirectNew:
        local: Optional["LocalAtDefineTime"]

        def __new__(cls, **kwargs):
            super().__new__(cls, **kwargs)

    _empty_callsite_scope(DirectNew, val)

    @record_custom()
    class InDirectNew:
        local: Optional["LocalAtDefineTime"]

        def __new__(cls, **kwargs):
            super().__new__(cls, **kwargs)

    _empty_callsite_scope(InDirectNew, val)


def test_didnt_override_new():
    with pytest.raises(check.CheckError):

        @record_custom()
        class Failed:
            local: str | None

    with pytest.raises(check.CheckError):

        @record_custom
        class FailedAgain:
            local: str | None


def test_empty():
    @record
    class Empty: ...

    assert Empty()


def test_optional_arg() -> None:
    @record
    class Opt:
        maybe: str | None = None
        always: str | None

    assert Opt(always="set")
    assert Opt(always="set", maybe="x").maybe == "x"

    @record(checked=False)
    class Other:
        maybe: str | None = None
        always: str | None

    assert Other(always="set")
    assert Other(always="set", maybe="x").maybe == "x"


def test_dont_share_containers() -> None:
    @record
    class Empties:
        items: list[str] = []
        map: dict[str, str] = {}

    e_1 = Empties()
    e_2 = Empties()
    assert e_1.items is not e_2.items
    assert e_1.map is not e_2.map


def test_sentinel():
    _unset = object()

    @record
    class Sample:
        val: Any | None = _unset

    assert Sample().val is _unset
    assert Sample(val=None).val is None

    @record(checked=False)
    class OtherSample:
        val: Any | None = _unset

    assert OtherSample().val is _unset
    assert OtherSample(val=None).val is None


@pytest.mark.parametrize(
    "fields, defaults, expected",
    [
        (
            ["name"],
            {},
            (
                ", *, name",
                "",
            ),
        ),
        # defaults dont need to be in certain order since we force kwargs
        # None handled directly by arg default
        (
            ["name", "age", "f"],
            {"age": None},
            (
                ", *, name, age = None, f",
                "",
            ),
        ),
        # empty container defaults get fresh copies via assignments
        (
            ["things"],
            {"things": []},
            (
                ", *, things = None",
                "things = things if things is not None else []",
            ),
        ),
        (
            ["map"],
            {"map": {}},
            (
                ", *, map = None",
                "map = map if map is not None else {}",
            ),
        ),
        # base case - default values resolved by reference to injected local
        (
            ["val"],
            {"val": object()},
            (
                f", *, val = {INJECTED_DEFAULT_VALS_LOCAL_VAR}['val']",
                "",
            ),
        ),
    ],
)
def test_build_args_and_assign(fields, defaults, expected):
    # tests / documents shared utility fn
    # don't hesitate to delete this upon refactor
    assert (
        build_args_and_assignment_strs(
            fields,
            defaults,
            kw_only=True,
        )
        == expected
    )


@record
class Person:
    name: str
    age: int


@record_custom
class Agent(IHaveNew):
    name: str
    secrets: list[str]

    def __new__(cls, name: str, **kwargs):
        return super().__new__(
            cls,
            name=name,
            secrets=kwargs.get("secrets", []),
        )


def test_pickle():
    p = Person(name="Lyra", age=2)
    assert p == pickle.loads(pickle.dumps(p))

    a = Agent(name="smith", secrets=["many"])
    assert a == pickle.loads(pickle.dumps(a))

    a2 = Agent(name="mr. clean")
    assert a2 == pickle.loads(pickle.dumps(a2))


def test_base_class_conflicts() -> None:
    class ConflictPropBase(ABC):
        @property
        def prop(self): ...

    with pytest.raises(check.CheckError, match="Conflicting non-abstract @property"):

        @record
        class X(ConflictPropBase):
            prop: Any

    class AbsPropBase(ABC):
        @property
        @abstractmethod
        def abstract_prop(self): ...

        @property
        @abstractmethod
        def abstract_prop_with_default(self) -> int: ...

    class DidntImpl(AbsPropBase): ...

    with pytest.raises(
        TypeError,
        match="Can't instantiate abstract class DidntImpl",
    ):
        DidntImpl()  # type: ignore # good job type checker

    @record
    class A(AbsPropBase):
        abstract_prop: Any
        abstract_prop_with_default: int = 0  # pyright: ignore[reportIncompatibleMethodOverride]

    assert A(abstract_prop=4).abstract_prop == 4
    assert A(abstract_prop=4).abstract_prop_with_default == 0

    class ConflictFnBase:
        def some_method(self): ...

    with pytest.raises(check.CheckError, match="Conflicting function"):

        @record
        class _(ConflictFnBase):
            some_method: Any

    with pytest.raises(check.CheckError, match="will have to override __new__"):

        def _some_func():
            return 4

        @record
        class _:
            thing: Any = _some_func


def test_lazy_import():
    @record
    class BadModel:
        foos: list["TestType"]

    with pytest.raises(check.CheckError, match="Unable to resolve"):
        BadModel(foos=[])

    @record
    class AnnotatedModel:
        foos: list[Annotated["TestType", ImportFrom("dagster_shared.utils.test")]]

    assert AnnotatedModel(foos=[])

    with pytest.raises(
        check.CheckError, match=r"Expected <class 'dagster_shared.utils.test.TestType'>"
    ):
        AnnotatedModel(foos=[1, 2, 3])  # pyright: ignore[reportArgumentType]

    def _out_of_scope():
        from dagster_shared.utils.test import TestType

        return AnnotatedModel(foos=[TestType()])

    assert _out_of_scope()


class Complex:
    def __init__(self, s: str):
        self.s = s


@record_custom(field_to_new_mapping={"foo_str": "foo"})
class Remapped(IHaveNew):
    foo_str: str

    def __new__(cls, foo: str | Complex):
        if isinstance(foo, Complex):
            foo = foo.s

        return super().__new__(
            cls,
            foo_str=foo,
        )

    @cached_property
    def foo(self) -> Complex:
        return Complex(self.foo_str)


def test_new_remaps_fields() -> None:
    r = Remapped(foo="test")
    assert r.foo.s == "test"
    r2 = copy(r)
    assert r2.foo.s == "test"

    assert pickle.loads(pickle.dumps(r)) == r


def test_docs():
    @record
    class Documented:
        """So much to know about this class."""

    assert Documented.__doc__


def test_make_hashable():
    @record
    class Nope:
        stuff: Sequence[Any]

    n = Nope(stuff=[1, 2, 3])
    with pytest.raises(TypeError, match="unhashable"):
        hash(n)

    @record
    class Yep:
        stuff: Sequence[Any]

        def __hash__(self):
            return hash_collection(self)  # pyright: ignore[reportArgumentType]

    y = Yep(stuff=[1, 2, 3])
    assert hash(y)
    assert hash(y) == hash(Yep(stuff=[1, 2, 3]))


def test_generic() -> None:
    T = TypeVar("T")

    @record
    class Things(Generic[T]):
        things: Sequence[T]

        @property
        def first_thing(self) -> T:
            return self.things[0]

    class StringThings(Things[str]): ...

    class IntThings(Things[int]): ...

    assert StringThings(things=["a", "b"]).first_thing == "a"
    assert IntThings(things=[1, 2]).first_thing == 1


def test_generic_nested() -> None:
    T = TypeVar("T")

    @whitelist_for_serdes
    @record
    class Things(Generic[T]):
        things: Sequence[T]

        @property
        def first_thing(self) -> T:
            return self.things[0]

    @whitelist_for_serdes
    @record
    class HolderOfThings(Generic[T]):
        things: Things[T]

    @record
    class HolderOfMaybeStrings:
        strings: Things[str | None]

    @record
    class MaybeHolderOfThings(Generic[T]):
        things: Things[T] | None

    HolderOfInts = HolderOfThings[int]

    assert HolderOfThings(things=Things(things=[1, 2])).things.first_thing == 1
    assert HolderOfMaybeStrings(strings=Things(things=["a", "b", None])).strings.first_thing == "a"
    assert HolderOfInts(things=Things(things=[1, 2])).things.first_thing == 1
    assert MaybeHolderOfThings(things=None).things is None

    holder_of_things = HolderOfThings(things=Things(things=[1, 2]))
    assert deserialize_value(serialize_value(holder_of_things)) == holder_of_things


def test_subclass_propagate_basic() -> None:
    @record
    class Super:
        some_var: int
        some_var_default: int = 1

    @record
    class Sub(Super):
        other_var: int
        other_var_default: int = 2

    default = Sub(some_var=1, other_var=2)
    assert default.some_var_default == 1
    assert default.other_var_default == 2

    non_default = Sub(some_var=1, other_var=2, some_var_default=3, other_var_default=4)
    assert non_default.some_var_default == 3
    assert non_default.other_var_default == 4

    with pytest.raises(TypeError, match="some_var"):
        assert Sub(other_var=2)  # type: ignore

    with pytest.raises(TypeError, match="other_var"):
        assert Sub(some_var=2)  # type: ignore


def test_subclass_propagate_change_defaults() -> None:
    @record
    class Super:
        a: int
        b: int = 1

    @record
    class Sub(Super):
        # add a default where none existed
        a: int = 0
        # change the default
        b: int = 0

        c: int
        d: int = 2

    sub = Sub(c=10)
    assert sub.a == 0
    assert sub.b == 0
    assert sub.c == 10
    assert sub.d == 2

    @record
    class SubSub(Sub):
        c: int = -1

    subsub = SubSub()
    assert subsub.c == -1
    assert subsub.b == 0

    assert repr(subsub) == "SubSub(a=0, b=0, c=-1, d=2)"


def test_generic_with_propagate() -> None:
    T = TypeVar("T")

    class Base(Generic[T]): ...

    @record
    class RecordBase(Base[T]):
        label: str | None = None

    @record
    class SubAdditionalArg(RecordBase):
        some_val: str

    obj = SubAdditionalArg(some_val="hi")
    assert SubAdditionalArg(some_val="hi").some_val == "hi"
    assert copy(obj, label="...").label == "..."
    assert copy(obj, some_val="new").some_val == "new"

    @record
    class SubAdditionalArgRecursive(RecordBase):
        vals: Sequence[RecordBase]

    obj = SubAdditionalArgRecursive(vals=[SubAdditionalArg(some_val="hi")])
    assert len(obj.vals) == 1
    assert copy(obj, label="...").label == "..."

    @record
    class SubAdditionalArgSpecific(RecordBase[int]):
        vals: Sequence[RecordBase[str]]

    obj = SubAdditionalArgSpecific(vals=[SubAdditionalArg(some_val="hi")])
    assert len(obj.vals) == 1
    assert copy(obj, label="...").label == "..."

    @record
    class SubAdditionalArgVariableBase(RecordBase[T]):
        vals: Sequence[RecordBase[T]]
        val: Base[T]

    obj = SubAdditionalArgVariableBase(
        vals=[SubAdditionalArg(some_val="hi")], val=SubAdditionalArg(some_val="bye")
    )
    assert len(obj.vals) == 1
    assert copy(obj, label="...").label == "..."

    @record
    class SubSubVariableBaseSpecific(SubAdditionalArgVariableBase[str]): ...

    obj = SubSubVariableBaseSpecific(
        vals=[SubAdditionalArg(some_val="hi")], val=SubAdditionalArg(some_val="bye")
    )
    assert len(obj.vals) == 1
    assert copy(obj, label="...").label == "..."

    @record
    class SubSubVariableBaseAny(SubAdditionalArgVariableBase): ...

    obj = SubSubVariableBaseAny(
        vals=[SubAdditionalArg(some_val="hi")], val=SubAdditionalArg(some_val="bye")
    )
    assert len(obj.vals) == 1
    assert copy(obj, label="...").label == "..."


def test_generic_with_propagate_type_checking() -> None:
    T = TypeVar("T")

    class Base(Generic[T]): ...

    @record
    class RecordBase(Base[T]):
        inner: T

    @record
    class SpecificRecord(RecordBase):
        val1: RecordBase
        val2: RecordBase[str]
        val3: Base
        val4: Base[int]

    valid_record = SpecificRecord(
        inner=...,
        val1=RecordBase(inner=1.23),
        val2=RecordBase(inner="hi"),
        val3=RecordBase(inner=1.23),
        val4=RecordBase(inner=1),
    )

    with pytest.raises(CheckError, match='"val1" is not a RecordBase'):
        copy(valid_record, val1=Base())

    with pytest.raises(CheckError, match='"val2" is not a RecordBase'):
        copy(valid_record, val2=Base())

    with pytest.raises(CheckError, match='"val3" is not a Base'):
        copy(valid_record, val3=3)

    with pytest.raises(CheckError, match='"val4" is not a Base'):
        copy(valid_record, val4=4)


def test_custom_subclass() -> None:
    @record_custom
    class Thing(IHaveNew):
        val: str

        def __new__(cls, val_short: str):
            return super().__new__(cls, val=val_short * 2)

    assert Thing(val_short="abc").val == "abcabc"

    with pytest.raises(
        CheckError,
        match=r"@record can not inherit from @record_custom",
    ):

        @record
        class SubThing(Thing):
            other_val: int


def test_replace() -> None:
    class NTExample(NamedTuple("_NT", [("name", str)])):
        def __new__(cls, name: str):
            check.invariant(name != "bad")
            return super().__new__(
                cls,
                name=name,
            )

    obj = NTExample("good")
    assert obj._asdict() == {"name": "good"}
    # namedtuple._replace bypasses NTExample.__new__
    assert obj._replace(name="bad")
    assert obj._replace(name="good") == obj
    assert obj._replace(name="foo").__class__ is obj.__class__

    @record_custom
    class Legacy(IHaveNew, LegacyNamedTupleMixin):
        name: str

        def __new__(cls, name: str):
            check.invariant(name != "bad")
            return super().__new__(
                cls,
                name=name,
            )

    obj = Legacy("good")
    assert obj._asdict() == {"name": "good"}
    # legacy _replace avoids __new__
    assert obj._replace(name="bad")
    assert obj._replace(name="good") == obj
    assert obj._replace(name="foo").__class__ is obj.__class__

    with pytest.raises(CheckError):
        obj._replace(name=4)

    @record
    class Parent:
        one: int

    @record
    class Child(Parent):
        two: int

        def boop(self):
            return "boop"

    obj = Child(one=1, two=2)
    replaced = replace(obj, two=2)
    assert replaced == obj
    replaced.boop()
    assert replaced.__class__ is obj.__class__

    with pytest.raises(CheckError):
        replace(obj, two="two")

    p = Parent(one=1)
    with pytest.raises(CheckError):
        replace(p, one="one")

    @record(checked=False)
    class Unchecked:
        name: str

    u = Unchecked(name=2)  # pyright: ignore[reportArgumentType]
    replace(u, name="unchecked")

    @record(checked=False)
    class UncheckedDefaults:
        name: str = "unchecked"
        age: int

    ud = UncheckedDefaults(age=2)
    replace(ud, name="steve")


def test_defensive_checks_running():
    # make sure we have enabled defensive checks in test, ideally as broadly as possible
    assert os.getenv("DAGSTER_RECORD_DEFENSIVE_CHECKS") == "true"


def test_allow_posargs():
    @record(kw_only=False)
    class Foo:
        a: int

    assert Foo(2)

    @record(kw_only=False)
    class Bar:
        a: int
        b: int
        c: int = 4

    assert Bar(1, 2)

    with pytest.raises(CheckError):

        @record(kw_only=False)
        class Baz:
            a: int = 4
            b: int  # type: ignore # good job type checker


def test_posargs_inherit():
    @record(kw_only=False)
    class Parent:
        name: str

    @record(kw_only=False)
    class Child(Parent):
        parent: Parent

    p = Parent("Alex")
    assert p
    c = Child("Lyra", p)
    assert c

    # test kw_only not being aligned
    with pytest.raises(CheckError):

        @record
        class Bad(Parent):
            other: str

    with pytest.raises(CheckError):

        @record
        class A:
            a: int

        @record(kw_only=False)
        class B(A):
            b: int


def test_pydantic() -> None:
    @record
    class Simple:
        a: str
        b: str

    class SimpleHolder(BaseModel):
        simple: Simple

    holder = SimpleHolder(simple=Simple(a="a", b="b"))
    assert TypeAdapter(SimpleHolder).validate_python(holder)

    @record_custom
    class Custom(IHaveNew):
        ab: str
        c: str

        def __new__(cls, a: str, b: str, c: int):
            return super().__new__(cls, ab=a + b, c=str(c))

    class CustomHolder(BaseModel):
        custom: Custom
        custom_list: Sequence[Custom]
        custom_mapping: Mapping[str, Custom]
        custom_list_mapping: Mapping[str, Sequence[Custom]]
        optional_custom_list_mapping: Mapping[str, Sequence[Custom]] | None

    c = Custom("a", "b", 1)
    holder = CustomHolder(
        custom=c,
        custom_list=[c],
        custom_mapping={"c": c},
        custom_list_mapping={"c": [c]},
        optional_custom_list_mapping={"c": [c]},
    )
    assert TypeAdapter(CustomHolder).validate_python(holder)


def test_runtime_typecheck_pydantic_field() -> None:
    with pytest.raises(
        CheckError, match=r"pydantic.Field is not supported as a default value for @record fields.*"
    ):

        @record
        class MyClass:
            foo: str = Field(default="foo")
            bar: int | None = Field(default=None)



"""Compressed serialization for serdes objects.

Provides deduplication and columnar packing of serialized data. Plugs a custom
object_handler into ``_transform_for_serialization`` so columnar collection
happens during the single serialization pass.

Columnar packing works by:

1. Grouping serialized objects by class, storing field names once per class
2. Storing each object's values as a compact positional row (no field names)
3. Deduplicating identical objects via content hashing (same-content rows
   share the same row index)
4. Replacing inline objects with compact ``{"__oid__": "tableId:rowIndex"}``
   references (numeric table IDs minimize payload size)

Usage::

    packed_json = serialize_deduped(cursor)
    cursor = deserialize_deduped(packed_json, as_type=AssetDaemonCursor)
"""

from collections.abc import Mapping
from typing import Any, cast, overload

from dagster_shared import seven
from dagster_shared.match import match_type
from dagster_shared.serdes.serdes import (
    _WHITELIST_MAP,
    JsonSerializableValue,
    PackableValue,
    SerializableObject,
    T_PackableValue,
    U_PackableValue,
    UnpackContext,
    UnpackedValue,
    WhitelistMap,
    _transform_for_serialization,
    _unpack_object,
)

# Wire format keys. Single-letter table keys minimize serialized size.
_COLUMNAR_KEY = "__columnar__"  # envelope marker: this is a columnar-packed payload
_OBJ_ID_KEY = "__oid__"  # inline ref: replaced object → "tableId:rowIndex"
_TABLE_CLASS_KEY = "c"  # table field: the serdes storage name for this class
_TABLE_FIELDS_KEY = "f"  # table field: ordered list of field names
_TABLE_ROWS_KEY = "r"  # table field: list of rows (each row is positional values)


###############################################################################
# Public API
###############################################################################


def serialize_deduped(
    val: PackableValue,
    whitelist_map: WhitelistMap = _WHITELIST_MAP,
) -> str:
    """Serialize a value into columnar-packed JSON with deduplication."""
    collector = _ColumnarCollector()
    handler = _make_collecting_handler(collector)

    tree = _transform_for_serialization(
        val,
        whitelist_map=whitelist_map,
        object_handler=handler,
        # Empty descent_path disables path tracking in error messages for perf.
        # Acceptable because callers (e.g. asset daemon) catch and re-raise with
        # their own context.
        descent_path="",
    )

    envelope = {
        _COLUMNAR_KEY: True,
        "tables": collector.tables,
        "value": tree,
    }
    return seven.json.dumps(envelope)


@overload
def deserialize_deduped(
    val: str,
    as_type: tuple[type[T_PackableValue], type[U_PackableValue]],
    whitelist_map: WhitelistMap = ...,
) -> T_PackableValue | U_PackableValue: ...


@overload
def deserialize_deduped(
    val: str,
    as_type: type[T_PackableValue],
    whitelist_map: WhitelistMap = ...,
) -> T_PackableValue: ...


@overload
def deserialize_deduped(
    val: str,
    as_type: None = ...,
    whitelist_map: WhitelistMap = ...,
) -> PackableValue: ...


def deserialize_deduped(
    val: str,
    as_type: type[T_PackableValue]
    | tuple[type[T_PackableValue], type[U_PackableValue]]
    | None = None,
    whitelist_map: WhitelistMap = _WHITELIST_MAP,
) -> PackableValue | T_PackableValue | U_PackableValue:
    """Deserialize a JSON string produced by ``serialize_deduped``."""
    parsed = seven.json.loads(val)

    context = UnpackContext()
    unpacked = _unpack_columnar(
        parsed["value"], parsed["tables"], whitelist_map, context, row_cache={}
    )
    result = context.finalize_unpack(unpacked)

    if as_type is not None and not match_type(result, as_type):  # type: ignore[arg-type]
        from dagster_shared.serdes.errors import DeserializationError

        raise DeserializationError(
            f"Unpacked object was not expected type {as_type}, got {type(result)}"
        )
    return result


###############################################################################
# Columnar packing (serialization)
###############################################################################


def _make_hashable(val: Any) -> Any:
    """Convert a JSON-serializable value to a hashable form for dedup.

    Fast-paths ``__oid__`` ref dicts (already simple strings) to avoid
    building unnecessary tuples — these make up ~30% of row values in
    practice.
    """
    if isinstance(val, dict):
        oid = val.get(_OBJ_ID_KEY)
        if oid is not None:
            return oid
        # Only bare mapping dicts reach here — @record instances are already
        # replaced with __oid__ refs (fast-pathed above) by the time _make_hashable runs.
        return tuple(sorted((k, _make_hashable(v)) for k, v in val.items()))
    if isinstance(val, list):
        return tuple(_make_hashable(v) for v in val)
    return val


class _ColumnarCollector:
    """Collects serialized objects into columnar tables during serialization.

    Each class gets a table with field names stored once and values as compact
    rows.  Duplicate objects (by value-equal row content) share the same row
    index.  Tables are stored as a list and referenced by numeric table_id in
    ``__oid__`` refs to minimize payload size.
    """

    def __init__(self) -> None:
        self.tables: list[dict[str, Any]] = []
        self._class_to_table_id: dict[str, int] = {}
        # (table_id, hashable_row) -> row index for dedup
        self._dedup: dict[tuple, int] = {}

    def register(
        self, class_name: str, fields_dict: Mapping[str, JsonSerializableValue]
    ) -> JsonSerializableValue:
        """Register a serialized object and return an ``__oid__`` ref.

        ``fields_dict`` contains already-serialized field values — nested
        objects are already ``__oid__`` refs, so hashing covers only one level.
        """
        if class_name not in self._class_to_table_id:
            table_id = len(self.tables)
            self._class_to_table_id[class_name] = table_id
            fields = sorted(k for k in fields_dict if k != "__class__")
            self.tables.append(
                {
                    _TABLE_CLASS_KEY: class_name,
                    _TABLE_FIELDS_KEY: fields,
                    _TABLE_ROWS_KEY: [],
                }
            )

        table_id = self._class_to_table_id[class_name]
        table = self.tables[table_id]
        row = [fields_dict.get(f) for f in table[_TABLE_FIELDS_KEY]]

        row_key = (table_id, _make_hashable(row))
        existing_idx = self._dedup.get(row_key)
        if existing_idx is not None:
            return {_OBJ_ID_KEY: f"{table_id}:{existing_idx}"}

        idx = len(table[_TABLE_ROWS_KEY])
        table[_TABLE_ROWS_KEY].append(row)
        self._dedup[row_key] = idx
        return {_OBJ_ID_KEY: f"{table_id}:{idx}"}


def _make_collecting_handler(collector: _ColumnarCollector):
    """Create an ``object_handler`` that collects objects into columnar tables.

    Plugs into ``_transform_for_serialization``'s ``object_handler`` parameter.
    Children are serialized first via ``pack_items`` (which calls back into this
    handler), so by the time we register a row, all nested objects are already
    ``__oid__`` refs.
    """

    def handler(
        obj: SerializableObject,
        wm: WhitelistMap,
        descent_path: str,
    ) -> JsonSerializableValue:
        klass_name = obj.__class__.__name__
        serializer = wm.object_serializers[klass_name]
        fields_dict = dict(serializer.pack_items(obj, wm, handler, descent_path))
        storage_name = cast("str", fields_dict["__class__"])
        return collector.register(storage_name, fields_dict)

    return handler


###############################################################################
# Columnar unpacking (deserialization)
###############################################################################


def _unpack_columnar(
    val: JsonSerializableValue,
    tables: list[dict[str, Any]],
    whitelist_map: WhitelistMap,
    context: UnpackContext,
    row_cache: dict[str, UnpackedValue],
) -> UnpackedValue:
    """Recursively unpack a columnar-packed value tree.

    Each ``__oid__`` ref is resolved exactly once and cached — subsequent
    references return the same Python object.
    """
    if isinstance(val, list):
        return [_unpack_columnar(item, tables, whitelist_map, context, row_cache) for item in val]

    if isinstance(val, dict):
        if _OBJ_ID_KEY in val:
            oid = cast("str", val[_OBJ_ID_KEY])
            if oid not in row_cache:
                table_id_str, idx_str = oid.rsplit(":", 1)
                table = tables[int(table_id_str)]
                row = table[_TABLE_ROWS_KEY][int(idx_str)]

                obj_dict: dict[str, Any] = {"__class__": table[_TABLE_CLASS_KEY]}
                for field_name, field_val in zip(table[_TABLE_FIELDS_KEY], row):
                    obj_dict[field_name] = _unpack_columnar(
                        field_val, tables, whitelist_map, context, row_cache
                    )
                row_cache[oid] = _unpack_object(obj_dict, whitelist_map, context)
            return row_cache[oid]

        # Non-columnar dict — a plain serdes object or raw mapping. Delegate to
        # _unpack_object which handles __class__ detection and passthrough.
        return _unpack_object(
            {
                k: _unpack_columnar(v, tables, whitelist_map, context, row_cache)
                for k, v in val.items()
            },
            whitelist_map,
            context,
        )

    return val



from dagster_shared.check.builder import (
    EvalContext as EvalContext,
    ImportFrom as ImportFrom,
    NoneType as NoneType,
    build_check_call_str as build_check_call_str,
)
from dagster_shared.check.functions import (
    CheckError as CheckError,
    ElementCheckError as ElementCheckError,
    NotImplementedCheckError as NotImplementedCheckError,
    Numeric as Numeric,
    ParameterCheckError as ParameterCheckError,
    assert_never as assert_never,
    bool_elem as bool_elem,
    bool_param as bool_param,
    callable_param as callable_param,
    class_param as class_param,
    dict_elem as dict_elem,
    dict_param as dict_param,
    failed as failed,
    float_elem as float_elem,
    float_param as float_param,
    generator as generator,
    generator_param as generator_param,
    inst as inst,
    inst_param as inst_param,
    int_elem as int_elem,
    int_param as int_param,
    int_value_param as int_value_param,
    invariant as invariant,
    is_callable as is_callable,
    is_dict as is_dict,
    is_iterable as is_iterable,
    is_list as is_list,
    is_tuple as is_tuple,
    iterable_param as iterable_param,
    iterator_param as iterator_param,
    list_elem as list_elem,
    list_param as list_param,
    literal_param as literal_param,
    mapping_param as mapping_param,
    not_implemented as not_implemented,
    not_none as not_none,
    not_none_param as not_none_param,
    numeric_param as numeric_param,
    opt_bool_elem as opt_bool_elem,
    opt_bool_param as opt_bool_param,
    opt_callable_param as opt_callable_param,
    opt_class_param as opt_class_param,
    opt_dict_elem as opt_dict_elem,
    opt_dict_param as opt_dict_param,
    opt_float_elem as opt_float_elem,
    opt_float_param as opt_float_param,
    opt_generator as opt_generator,
    opt_generator_param as opt_generator_param,
    opt_inst as opt_inst,
    opt_inst_param as opt_inst_param,
    opt_int_elem as opt_int_elem,
    opt_int_param as opt_int_param,
    opt_iterable_param as opt_iterable_param,
    opt_list_elem as opt_list_elem,
    opt_list_param as opt_list_param,
    opt_literal_param as opt_literal_param,
    opt_mapping_param as opt_mapping_param,
    opt_nonempty_str_param as opt_nonempty_str_param,
    opt_nullable_dict_elem as opt_nullable_dict_elem,
    opt_nullable_dict_param as opt_nullable_dict_param,
    opt_nullable_iterable_param as opt_nullable_iterable_param,
    opt_nullable_list_param as opt_nullable_list_param,
    opt_nullable_mapping_param as opt_nullable_mapping_param,
    opt_nullable_sequence_param as opt_nullable_sequence_param,
    opt_nullable_set_param as opt_nullable_set_param,
    opt_nullable_tuple_param as opt_nullable_tuple_param,
    opt_numeric_param as opt_numeric_param,
    opt_path_param as opt_path_param,
    opt_sequence_param as opt_sequence_param,
    opt_set_param as opt_set_param,
    opt_str_elem as opt_str_elem,
    opt_str_param as opt_str_param,
    opt_tuple_elem as opt_tuple_elem,
    opt_tuple_param as opt_tuple_param,
    opt_two_dim_dict_param as opt_two_dim_dict_param,
    param_invariant as param_invariant,
    path_param as path_param,
    sequence_param as sequence_param,
    set_param as set_param,
    str_elem as str_elem,
    str_param as str_param,
    tuple_elem as tuple_elem,
    tuple_param as tuple_param,
    two_dim_dict_param as two_dim_dict_param,
    two_dim_list_param as two_dim_list_param,
    two_dim_mapping_param as two_dim_mapping_param,
)



"""Internal py2/3 compatibility library. A little more than six."""

import importlib.util
import inspect
import os
import pkgutil
import shlex
import sys
import time
from collections.abc import Sequence
from importlib.machinery import ModuleSpec
from pathlib import Path
from types import ModuleType
from typing import Any, Callable, List, Optional, Type, TypeGuard, Union  # noqa: F401, UP035

import dagster_shared.seven.json as json  # noqa: F401
from dagster_shared.seven.json import (
    JSONDecodeError as JSONDecodeError,
    dump as dump,
    dumps as dumps,
)
from dagster_shared.seven.temp_dir import get_system_temp_directory as get_system_temp_directory

IS_WINDOWS = os.name == "nt"
IS_PYTHON_3_12 = (sys.version_info[0], sys.version_info[1]) == (3, 12)
IS_PYTHON_3_13 = (sys.version_info[0], sys.version_info[1]) == (3, 13)
IS_PYTHON_3_14 = (sys.version_info[0], sys.version_info[1]) == (3, 14)

# TODO implement a generic import by name -- see https://stackoverflow.com/questions/301134/how-to-import-a-module-given-its-name


# https://stackoverflow.com/a/67692/324449
def import_module_from_path(module_name: str, path_to_file: str | Path) -> ModuleType:
    import importlib.util

    spec = importlib.util.spec_from_file_location(module_name, path_to_file)
    if spec is None:
        raise Exception(
            f"Can not import module {module_name} from path {path_to_file}, unable to load spec."
        )

    if sys.modules.get(spec.name) and spec.origin:
        f = sys.modules[spec.name].__file__
        # __file__ can be relative depending on current working directory
        if f and os.path.abspath(f) == os.path.abspath(spec.origin):
            return sys.modules[spec.name]

    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    assert spec.loader
    spec.loader.exec_module(module)
    return module


def import_uncached_module_from_path(module_name: str, path_to_file: str) -> ModuleType:
    import importlib.util

    spec = importlib.util.spec_from_file_location(module_name, path_to_file)
    if spec is None:
        raise Exception(
            f"Can not import module {module_name} from path {path_to_file}, unable to load spec."
        )

    module = importlib.util.module_from_spec(spec)
    assert spec.loader
    spec.loader.exec_module(module)
    return module


time_fn = time.perf_counter


def get_arg_names(callable_: Callable[..., Any]) -> Sequence[str]:
    return [
        parameter.name
        for parameter in inspect.signature(callable_).parameters.values()
        if parameter.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
    ]


# https://stackoverflow.com/a/58437485/324449
def is_module_available(module_name: str) -> bool:
    # python 3.4 and above
    import importlib.util

    loader = importlib.util.find_spec(module_name)

    return loader is not None


def is_lambda(target: object) -> TypeGuard[Callable[..., Any]]:
    return callable(target) and getattr(target, "__name__", None) == "<lambda>"


def is_function_or_decorator_instance_of(
    target: object, kls: type[Any]
) -> TypeGuard[Callable[..., Any]]:
    return inspect.isfunction(target) or (isinstance(target, kls) and hasattr(target, "__name__"))


def qualname_differs(target: object) -> bool:
    return hasattr(target, "__qualname__") and getattr(target, "__qualname__") != getattr(
        target, "__name__"
    )


def xplat_shlex_split(s: str) -> list[str]:
    if IS_WINDOWS:
        return shlex.split(s, posix=False)

    return shlex.split(s)


def get_import_error_message(import_error: ImportError) -> str:
    return import_error.msg


def is_subclass(child_type: type[Any], parent_type: type[Any]):
    """Due to some pathological interactions betwen bugs in the Python typing library
    (https://github.com/python/cpython/issues/88459 and
    https://github.com/python/cpython/issues/89010), some types (list[str] in Python 3.9, for
    example) pass inspect.isclass check above but then raise an exception if issubclass is called
    with the same class. This function provides a workaround for that issue.
    """
    if not inspect.isclass(child_type):
        return False

    try:
        return issubclass(child_type, parent_type)
    except TypeError:
        return False


def resolve_module_pattern(pattern: str) -> list[str]:
    """Find all modules that match the given pattern segments.

    Note that this will only return modules that have been verified to exist using
    importlib.util.find_spec. This means that a pure literal pattern like "foo.bar" will
    return an empty list if the module does not exist.

    Args:
        pattern_segments: List of pattern segments, where * matches any segment name.

    Returns:
        List of matching module names.
    """
    pattern_segments = pattern.split(".")
    return _gather_modules([], pattern_segments)


def _gather_modules(current_segments: list[str], remaining_pattern: list[str]) -> list[str]:
    if not remaining_pattern:
        # We've matched the full pattern, return the current module path
        module_name = ".".join(current_segments)
        # Verify the module actually exists and can be imported
        if _get_module_spec(module_name) is not None:
            return [module_name]
        else:
            return []

    current_pattern = remaining_pattern[0]
    rest_pattern = remaining_pattern[1:]

    # Wildcard
    if current_pattern == "*":
        if len(current_segments) == 0:  # top-level wildcard
            current_module_path = None
        else:
            current_module_name = ".".join(current_segments)
            current_module_spec = _get_module_spec(current_module_name)
            if (
                current_module_spec is None
                or current_module_spec.submodule_search_locations is None
            ):
                return []
            else:
                current_module_path = current_module_spec.submodule_search_locations

        # Explore all top-level packages
        matches = []
        for _, name, _ in pkgutil.iter_modules(current_module_path):
            matches.extend(_gather_modules([*current_segments, name], rest_pattern))
        return matches

    # Literal
    else:
        # Literal segment - add it and continue
        return _gather_modules([*current_segments, current_pattern], rest_pattern)


def _get_module_spec(module_name: str) -> ModuleSpec | None:
    try:
        return importlib.util.find_spec(module_name)

    # ModuleNotFoundError gets thrown if you try to find a spec where one of the parent modules does
    # not exist, e.g. "foo.bar.baz" when "foo" does not exist.
    except ModuleNotFoundError:
        return None


def match_module_pattern(module_name: str, pattern: str) -> bool:
    """Check if a module name matches a given pattern."""
    pattern_segments = pattern.split(".")
    module_segments = module_name.split(".")

    if len(pattern_segments) != len(module_segments):
        return False

    for pattern_segment, module_segment in zip(pattern_segments, module_segments):
        if pattern_segment != "*" and pattern_segment != module_segment:
            return False

    return True


def is_valid_module_pattern(pattern: str) -> bool:
    """Check if a module pattern is valid."""
    if not pattern:
        return False

    for segment in pattern.split("."):
        if not (segment.isidentifier() or segment == "*"):
            return False

    return True


def load_module_object(module_name: str, attr: str) -> object:
    """Loads a top-level attribute from a module.

    Args:
        module_name: The name of the module to load the attribute from.
        attr: The name of the attribute to load.

    Returns:
        The attribute value.

    Raises:
        DagsterUnresolvableSymbolError: If the module or attribute is not found.
    """
    from dagster_shared.error import DagsterUnresolvableSymbolError

    try:
        module = importlib.import_module(module_name)
        if not hasattr(module, attr):
            raise DagsterUnresolvableSymbolError(
                f"Module `{module_name}` has no attribute `{attr}`."
            )
        return getattr(module, attr)
    except ModuleNotFoundError as e:
        raise DagsterUnresolvableSymbolError(f"Module `{module_name}` not found.") from e
    except ImportError as e:
        raise DagsterUnresolvableSymbolError(f"Error loading module `{module_name}`.") from e



import os
import sys
from collections.abc import Callable, Mapping
from pathlib import Path
from typing import Any, Final, Literal, TypeAlias


def is_windows() -> bool:
    return sys.platform == "win32"


def get_default_dg_user_config_path() -> Path:
    if is_windows():
        return Path.home() / "AppData" / "dg" / "dg.toml"
    else:
        config_home = Path(os.getenv("XDG_CONFIG_HOME", os.path.expanduser("~/.config")))
        return config_home / "dg.toml"


def load_config(config_path: Path) -> dict[str, Any]:
    import tomlkit
    import yaml

    if config_path.suffix == ".toml":
        return tomlkit.parse(config_path.read_text()).unwrap()
    else:
        return yaml.safe_load(config_path.read_text()) or {}


def write_config(config_path: Path, config: dict[str, Any]):
    import tomlkit
    import yaml

    if config_path.suffix == ".toml":
        with open(config_path, "w") as f:
            tomlkit.dump(config, f)
    else:
        with open(config_path, "w") as f:
            yaml.safe_dump(config, f)


def get_dg_config_path() -> Path:
    return Path(os.getenv("DG_CLI_CONFIG", get_default_dg_user_config_path()))


def does_dg_config_file_exist() -> bool:
    return get_dg_config_path().exists()


# The format determines whether settings are nested under the `tool.dg` section
# (`pyproject.toml`) or not (`dg.toml`).
DgConfigFileFormat: TypeAlias = Literal["root", "nested"]
_DEFAULT_PROJECT_DEFS_SUBMODULE: Final = "defs"


def detect_dg_config_file_format(path: Path) -> DgConfigFileFormat:
    """Check if the file is a dg-specific toml file."""
    return "root" if path.name == "dg.toml" or path.name == ".dg.toml" else "nested"


def load_toml_as_dict(path: Path) -> dict[str, Any]:
    import tomlkit

    return tomlkit.parse(path.read_text()).unwrap()


def has_dg_file_config(
    path: Path, predicate: Callable[[Mapping[str, Any]], bool] | None = None
) -> bool:
    toml = load_toml_as_dict(path)
    # `dg.toml` is a special case where settings are defined at the top level
    if detect_dg_config_file_format(path) == "root":
        node = toml
    else:
        if "dg" not in toml.get("tool", {}):
            return False
        node = toml["tool"]["dg"]
    return predicate(node) if predicate else True


# NOTE: The presence of dg.toml will cause pyproject.toml to be ignored for purposes of dg config.
def discover_config_file(
    path: Path,
    predicate: Callable[[Mapping[str, Any]], bool] | None = None,
) -> Path | None:
    current_path = path.absolute()
    while True:
        config_file = locate_dg_config_in_folder(current_path, predicate)
        if config_file:
            return config_file
        if current_path == current_path.parent:  # root
            return None
        current_path = current_path.parent


def locate_dg_config_in_folder(
    path: Path,
    predicate: Callable[[Mapping[str, Any]], bool] | None = None,
) -> Path | None:
    current_path = path.absolute()
    dg_toml_path = current_path / "dg.toml"
    pyproject_toml_path = current_path / "pyproject.toml"
    if dg_toml_path.exists() and has_dg_file_config(dg_toml_path, predicate):
        return dg_toml_path
    elif pyproject_toml_path.exists() and has_dg_file_config(pyproject_toml_path, predicate):
        return pyproject_toml_path
    return None


def get_canonical_defs_module_name(defs_module_name: str | None, root_module_name: str) -> str:
    return defs_module_name or f"{root_module_name}.{_DEFAULT_PROJECT_DEFS_SUBMODULE}"
